
// hash and length combined mean its very hard to change simple editor blocks file, so simple editor players can only use those blocks
string simpleEditorHash = '4721fcd70890792a4800ea8f1f8134d60565a596da8af6aa3b0131d4141f1337';
int simpleEditorLength = 27161;
array<Block@> Blocks;

array<array<uint>> StraightBlocks = {}; 
array<array<uint>> StraightBlocks_up = {}; 
array<array<uint>> StraightBlocks_down = {}; 

void SetupStraightBlocks() {
    for (uint i = 0; i < Connections::PlatformAny; i++) {
        StraightBlocks.InsertLast({});
        StraightBlocks_up.InsertLast({});
        StraightBlocks_down.InsertLast({});
    }
}

enum LoadStatus {
    NotLoaded,
    ToLoadAll,
    LoadingAll,
    All,
    ToLoadSimpleEditor,
    LoadingSimpleEditor,
    SimpleEditor,
    LoadingExtra,
    ToLoadNonzeroWeight,
    LoadingNonzeroWeight,
    NonzeroWeight,
}

LoadStatus blocksStatus = LoadStatus::NotLoaded;

array<Tags> toTagArray(Json::Value tags) {
    array<Tags> result = {};
    result.Reserve(tags.Length);
    for (uint i = 0; i < tags.Length; i++) {
        result.InsertLast(Tags(uint(tags[i])));
    }
    return result;
}
array<int3> toint3Array(Json::Value int3s) {
    array<int3> result = {};
    result.Reserve(int3s.Length);
    for (uint i = 0; i < int3s.Length; i++) {
        result.InsertLast(int3(int(int3s[i][0]),int(int3s[i][1]),int(int3s[i][2])));
    }
    return result;
}

void LoadBlocks() {
    if (!Permissions::OpenAdvancedMapEditor()) {
        UI::ShowNotification(pluginName, "Error - trying to load advanced editor blocks without permission", notificationErrorColor);
        blocksStatus = LoadStatus::NotLoaded;
        return;
    }
    blocksStatus = LoadStatus::LoadingAll;
    Blocks = {};
    Json::Value JsonBlocks = Json::FromFile('BlockData.json');
    Blocks.Reserve(JsonBlocks.Length);
    Block@ b;
    SetupStraightBlocks();
    for (uint i = 0; i < JsonBlocks.Length; i++) {
        if (Time::get_Now() - GlobalFrameStartTime > 20) {
            yield();
            GlobalFrameStartTime = Time::get_Now();
        }
        Json::Value CurrentBlock = JsonBlocks[i];
        @b = Block(
            CurrentBlock.Get("NameID"),
            Connections(uint(CurrentBlock.Get("startConnector"))),
            Connections(uint(CurrentBlock.Get("endConnector"))),
            int3(CurrentBlock.Get("endPosition")[0], CurrentBlock.Get("endPosition")[1], CurrentBlock.Get("endPosition")[2]),
            Directions(uint(CurrentBlock.Get("Direction"))),
            toTagArray(CurrentBlock.Get("tags")),
            uint(CurrentBlock.Get("length")),
            Directions(uint(CurrentBlock.Get("RotationOffset"))),
            toint3Array(CurrentBlock.Get("PositionOffset")));
        Blocks.InsertLast(b);
        // if (string(CurrentBlock.Get("NameID")).StartsWith("Macroblock:Stadium\\Macroblocks\\LightSculpture")) {
        if (b.startConnector == b.endConnector && b.Direction == Directions::Forwards && b.endPosition == int3(0,0,1)) {
            StraightBlocks[b.startConnector].InsertLast(i);
        } else if (b.startConnector == b.endConnector && b.Direction == Directions::Forwards && b.endPosition == int3(0,1,1)) {
            StraightBlocks_up[b.startConnector].InsertLast(i);
        } else if (b.startConnector == b.endConnector && b.Direction == Directions::Forwards && b.endPosition == int3(0,-1,1)) {
            StraightBlocks_down[b.startConnector].InsertLast(i);
        }
    }
    blocksStatus = LoadStatus::All;
    if (debugPrint) {print('Loaded ' + Blocks.Length + ' blocks');}
#if RECREATEBLOCKSFILE
    CreateBlocks();
#endif
}

void LoadBlocksNonzeroWeight() {
    if (!Permissions::OpenAdvancedMapEditor()) {
        UI::ShowNotification(pluginName, "Error - trying to load advanced editor blocks without permission", notificationErrorColor);
        blocksStatus = LoadStatus::NotLoaded;
        return;
    }
    blocksStatus = LoadStatus::LoadingNonzeroWeight;
    Blocks = {};
    Json::Value JsonBlocks = Json::FromFile('BlockData.json');
    Blocks.Reserve(JsonBlocks.Length);
    Block@ b;
    array<Tags> tags;
    bool iszero;
    SetupStraightBlocks();
    uint blockIdx = 0; // index of block in block list; i is in currentBlocks which is now different
    for (uint i = 0; i < JsonBlocks.Length; i++) {
        if (Time::get_Now() - GlobalFrameStartTime > 20) {
            yield();
            GlobalFrameStartTime = Time::get_Now();
        }
        Json::Value CurrentBlock = JsonBlocks[i];
        tags = toTagArray(CurrentBlock.Get("tags"));
        iszero = false;
        for (uint j = 0; j < tags.Length; j++){
            if (getWeightFromTag(tags[j]) == 0) {
                iszero = true;
            }
        }
        if (iszero) {continue;}
        @b = Block(
            CurrentBlock.Get("NameID"),
            Connections(uint(CurrentBlock.Get("startConnector"))),
            Connections(uint(CurrentBlock.Get("endConnector"))),
            int3(CurrentBlock.Get("endPosition")[0], CurrentBlock.Get("endPosition")[1], CurrentBlock.Get("endPosition")[2]),
            Directions(uint(CurrentBlock.Get("Direction"))),
            tags,
            uint(CurrentBlock.Get("length")),
            Directions(uint(CurrentBlock.Get("RotationOffset"))),
            toint3Array(CurrentBlock.Get("PositionOffset")));
        Blocks.InsertLast(b);
        blockIdx++;
        if (b.startConnector == b.endConnector && b.Direction == Directions::Forwards && b.endPosition == int3(0,0,1)) {
            StraightBlocks[b.startConnector].InsertLast(blockIdx);
        } else if (b.startConnector == b.endConnector && b.Direction == Directions::Forwards && b.endPosition == int3(0,1,1)) {
            StraightBlocks_up[b.startConnector].InsertLast(blockIdx);
        } else if (b.startConnector == b.endConnector && b.Direction == Directions::Forwards && b.endPosition == int3(0,-1,1)) {
            StraightBlocks_down[b.startConnector].InsertLast(blockIdx);
        }
    }
    blocksStatus = LoadStatus::NonzeroWeight;
    if (debugPrint) {print('Loaded ' + Blocks.Length + ' blocks');}
}

void LoadBlocksSimpleEditor() {
    blocksStatus = LoadStatus::LoadingSimpleEditor;
    Blocks = {};
    IO::FileSource simpleEditorFile('BlockDataSimpleEditor.json');
    string fileContents = simpleEditorFile.ReadToEnd();
    if (fileContents != '') {
        string hash = Crypto::Sha256(fileContents);
        if (fileContents.Length != simpleEditorLength || hash != simpleEditorHash) {
            if (Permissions::OpenAdvancedMapEditor()) {
                UI::ShowNotification(pluginName, "Warning - simple editor blocks file is modified but you have advanced editor", notificationWarningColor);
                print("Warning - simple editor blocks file is modified but you have advanced editor access");
                print(fileContents.Length);
                print(hash);
            } else {
                blocksStatus = LoadStatus::NotLoaded;
                UI::ShowNotification(pluginName, "Couldn't load simple editor blocks - file is modified", notificationWarningColor);
                print("Couldn't load simple editor blocks - file is modified");
                return;
            }
        }
    }
    Json::Value JsonBlocks = Json::FromFile('BlockDataSimpleEditor.json');
    Blocks.Reserve(JsonBlocks.Length);
    Block@ b;
    SetupStraightBlocks();
    if (!(JsonBlocks.GetType() == Json::Type::Null)) {
        Blocks.Reserve(JsonBlocks.Length);
        for (uint i = 0; i < JsonBlocks.Length; i++) {
            if (Time::get_Now() - GlobalFrameStartTime > 20) {
                yield();
                GlobalFrameStartTime = Time::get_Now();
            }
            Json::Value CurrentBlock = JsonBlocks[i];
            @b = Block(
                CurrentBlock.Get("NameID"),
                Connections(uint(CurrentBlock.Get("startConnector"))),
                Connections(uint(CurrentBlock.Get("endConnector"))),
                int3(CurrentBlock.Get("endPosition")[0], CurrentBlock.Get("endPosition")[1], CurrentBlock.Get("endPosition")[2]),
                Directions(uint(CurrentBlock.Get("Direction"))),
                toTagArray(CurrentBlock.Get("tags")),
                uint(CurrentBlock.Get("length")),
                Directions(uint(CurrentBlock.Get("RotationOffset"))),
                toint3Array(CurrentBlock.Get("PositionOffset")));
            Blocks.InsertLast(b);
            if (b.startConnector == b.endConnector && b.Direction == Directions::Forwards && b.endPosition == int3(0,0,1)) {
                StraightBlocks[b.startConnector].InsertLast(i);
            } else if (b.startConnector == b.endConnector && b.Direction == Directions::Forwards && b.endPosition == int3(0,1,1)) {
                StraightBlocks_up[b.startConnector].InsertLast(i);
            } else if (b.startConnector == b.endConnector && b.Direction == Directions::Forwards && b.endPosition == int3(0,-1,1)) {
                StraightBlocks_down[b.startConnector].InsertLast(i);
            }
        }
        blocksStatus = LoadStatus::SimpleEditor;
        // CreateBlocksSimpleEditor();
        if (debugPrint) {print('Loaded ' + Blocks.Length + ' blocks');}
    } else {
        print('Failed to load simple editor blocks');
                UI::ShowNotification(pluginName, "Failed to load simple editor blocks", notificationErrorColor);
        blocksStatus = LoadStatus::NotLoaded;
    }
}


// load blocks from an extra file
void LoadExtraBlocks(const string &in filename) {
    if (!Permissions::OpenAdvancedMapEditor()) {
        UI::ShowNotification(pluginName, "Error - trying to load extra blocks without advanced editor permission", notificationErrorColor);
        blocksStatus = LoadStatus::NotLoaded;
        return;
    }
    blocksStatus = LoadStatus::LoadingExtra;
    Json::Value JsonBlocks = Json::FromFile(filename);
    if (JsonBlocks.GetType() == Json::Type::Null) {
        UI::ShowNotification(pluginName, "Couldn't read file", notificationErrorColor);
        print(filename);
        return;
    }
    Blocks.Reserve(Blocks.Length + JsonBlocks.Length);
    Block@ b;
    SetupStraightBlocks();
    for (uint i = 0; i < JsonBlocks.Length; i++) {
        if (Time::get_Now() - GlobalFrameStartTime > 20) {
            yield();
            GlobalFrameStartTime = Time::get_Now();
        }
        Json::Value CurrentBlock = JsonBlocks[i];
        @b = Block(
            CurrentBlock.Get("NameID"),
            Connections(uint(CurrentBlock.Get("startConnector"))),
            Connections(uint(CurrentBlock.Get("endConnector"))),
            int3(CurrentBlock.Get("endPosition")[0], CurrentBlock.Get("endPosition")[1], CurrentBlock.Get("endPosition")[2]),
            Directions(uint(CurrentBlock.Get("Direction"))),
            toTagArray(CurrentBlock.Get("tags")),
            uint(CurrentBlock.Get("length")),
            Directions(uint(CurrentBlock.Get("RotationOffset"))),
            toint3Array(CurrentBlock.Get("PositionOffset")));
        Blocks.InsertLast(b);
        // if (string(CurrentBlock.Get("NameID")).StartsWith("Macroblock:Stadium\\Macroblocks\\LightSculpture")) {
        if (b.startConnector == b.endConnector && b.Direction == Directions::Forwards && b.endPosition == int3(0,0,1)) {
            StraightBlocks[b.startConnector].InsertLast(i);
        } else if (b.startConnector == b.endConnector && b.Direction == Directions::Forwards && b.endPosition == int3(0,1,1)) {
            StraightBlocks_up[b.startConnector].InsertLast(i);
        } else if (b.startConnector == b.endConnector && b.Direction == Directions::Forwards && b.endPosition == int3(0,-1,1)) {
            StraightBlocks_down[b.startConnector].InsertLast(i);
        }
    }
    blocksStatus = LoadStatus::All;
    if (debugPrint) {print('Loaded ' + Blocks.Length + ' blocks');}
}
